from abc import ABC
from typing import Generic, Optional, Self

from ..with_base.types import T_in, T_out
from .pipelineable_facade import TPipelineable
from ..with_base.stage_callable import StageCallableType
from .pipeline_interface_mixin import (
    PipelineInterfaceMixin as ImplementsPipelineInterface,
)
from ..with_pipeline_processor.processor_interface import ProcessorInterface


class PipelineMixin(
    ImplementsPipelineInterface[T_in, T_out],
    Generic[TPipelineable, T_in, T_out],
    ABC,
):
    pipelineable_class: type[TPipelineable]

    def add(self, item: StageCallableType) -> Self:
        return self.add_item(item)

    def build(
        self,
        processor: Optional[ProcessorInterface[T_in, T_out]] = None,
    ) -> TPipelineable:
        return self.pipelineable_class(
            stage_instances=self.get_items(),
            processor=processor,
        )



from ...classes.pipeline import Pipeline
from ...classes.types import T_in, T_out
from .stateful_chained_processor import StatefulChainedProcessor


class StatefulChainedPipeline(Pipeline[T_in, T_out]):
    processor_class = StatefulChainedProcessor



from abc import ABC
from typing import Any, Generic, Optional

from .command_interface import CommandInterface
from .processor_interface_mixin import (
    ProcessorInterfaceMixin as ImplementsProcessorInterface,
)
from ..with_base.types import T_in, T_out
from ..with_base.stage_callable import StageCallableType


class ProcessorMixin(
    ImplementsProcessorInterface[T_in, T_out],
    Generic[T_in, T_out],
    ABC,
):
    command_class: Optional[type[CommandInterface[T_in, T_out]]] = None

    def __init__(
        self,
        command_class: Optional[type[CommandInterface[T_in, T_out]]] = None,
        *args: Any,
        **kwds: Any,
    ) -> None:
        super().__init__(*args, **kwds)  # type: ignore

        self.command_class = command_class or self.command_class

    def _make_command(
        self,
        payload: T_in,
        stages: list[StageCallableType],
        *args: Any,
        **kwds: Any,
    ) -> CommandInterface[T_in, T_out]:
        if self.command_class is None:
            raise ValueError("Command class not set")

        return self.command_class(
            processor=self, payload=payload, stages=stages, *args, **kwds
        )

    async def process(
        self,
        payload: T_in,
        stages: list[StageCallableType],
        *args: Any,
        **kwds: Any,
    ) -> T_out:
        command = self._make_command(payload, stages, *args, **kwds)

        return await command.execute()



from typing import Any, Protocol, TypeVar

from ..with_base.types import T_in, T_out
from ..with_pipeline_processor.processor_interface import (
    ProcessorInterface as WithProcessorBaseInterface,
)


class ProcessorFacade(
    WithProcessorBaseInterface[T_in, T_out],
    Protocol[T_in, T_out],
):
    pass


TProcessor = TypeVar("TProcessor", bound=ProcessorFacade[Any, Any])



from typing import Any, Optional, Protocol

from .command_interface import CommandInterface
from ..with_base.types import T_in, T_out
from ..with_base.stage_callable import StageCallableType
from ..with_pipeline_processor.processor_interface import (
    ProcessorInterface as WithProcessorBaseInterface,
)


class ProcessorInterfaceMixin(
    WithProcessorBaseInterface[T_in, T_out],
    Protocol[T_in, T_out],
):
    def __init__(
        self,
        command_class: Optional[type[CommandInterface[T_in, T_out]]] = None,
        *args: Any,
        **kwds: Any,
    ) -> None: ...

    def _make_command(
        self,
        payload: T_in,
        stages: list[StageCallableType],
        *args: Any,
        **kwds: Any,
    ) -> CommandInterface[T_in, T_out]: ...



import pytest
from thecodecrate_pipeline import (
    ChainedPipeline,
    ChainedProcessor,
    InterruptiblePipeline,
    InterruptibleProcessor,
    StatefulChainedPipeline,
    StatefulChainedProcessor,
)


@pytest.mark.asyncio
async def test_chained_pipeline():
    pipeline = (
        ChainedPipeline[int]()
        .pipe(lambda payload: payload + 1)
        .pipe(lambda payload: payload * 2)
    )
    assert await pipeline.process(5) == 12


@pytest.mark.asyncio
async def test_chained_processor():
    processor = ChainedProcessor[int]()

    result = await processor.process(
        payload=5,
        stages=[
            lambda payload: payload + 1,
            lambda payload: payload * 2,
        ],
    )
    assert result == 12

    result = await processor.process(
        payload=5,
        stages=[
            lambda payload: payload + 1,
        ],
    )
    assert result == 6


@pytest.mark.asyncio
async def test_interruptible_pipeline():
    pipeline = (
        InterruptiblePipeline[int](lambda payload: payload < 10)
        .pipe(lambda payload: payload + 2)
        .pipe(lambda payload: payload * 10)
        .pipe(lambda payload: payload * 10)
    )
    assert await pipeline.process(5) == 70


@pytest.mark.asyncio
async def test_interruptible_processor():
    processor = InterruptibleProcessor[int](lambda payload: payload < 10)

    result = await processor.process(
        payload=5,
        stages=[
            lambda payload: payload + 2,
            lambda payload: payload * 10,
            lambda payload: payload * 10,
        ],
    )
    assert result == 70

    result = await processor.process(
        payload=5,
        stages=[
            lambda payload: payload + 2,
        ],
    )
    assert result == 7


@pytest.mark.asyncio
async def test_stateful_chained_pipeline():
    pipeline = (
        StatefulChainedPipeline[int]()
        .pipe(lambda payload: payload + 1)
        .pipe(lambda payload: payload * 2)
    )
    assert await pipeline.process(5) == 12


@pytest.mark.asyncio
async def test_stateful_chained_processor():
    processor = StatefulChainedProcessor[int]()

    result = await processor.process(
        payload=5,
        stages=[
            lambda payload: payload + 1,
            lambda payload: payload * 2,
        ],
    )
    assert result == 12

    result = await processor.process(
        payload=5,
        stages=[
            lambda payload: payload + 1,
        ],
    )
    assert result == 6



from typing import Any, Protocol, TypeVar

from ..with_base.types import T_in, T_out
from ..with_processor_commands.command_interface import (
    CommandInterface as WithCommandBaseInterface,
)


class CommandFacade(
    WithCommandBaseInterface[T_in, T_out],
    Protocol[T_in, T_out],
):
    pass


TCommand = TypeVar(
    "TCommand", bound=CommandFacade[Any, Any], infer_variance=True
)



from abc import ABC
from typing import Any, Generic, Optional

from .command_facade import TCommand
from .processor_facade import ProcessorFacade
from .pipeline_interface_mixin import (
    PipelineInterfaceMixin as ImplementsPipelineInterface,
)
from ..with_base.types import T_in, T_out


class PipelineMixin(
    ImplementsPipelineInterface[TCommand, T_in, T_out],
    Generic[TCommand, T_in, T_out],
    ABC,
):
    base_processor_class: Optional[type[ProcessorFacade]] = None
    command_class: Optional[type[TCommand]] = None

    def __init__(self, *args: Any, **kwds: Any) -> None:
        super().__init__(*args, **kwds)  # type: ignore

        if self._should_make_command_processor():
            self.processor = self._make_command_processor()

    def _should_make_command_processor(self) -> bool:
        return (
            self.command_class is not None
            and self.base_processor_class is not None
        )

    def _make_command_processor(self) -> ProcessorFacade:
        if self.base_processor_class is None:
            raise ValueError("Base processor class not set")

        return self.base_processor_class(command_class=self.command_class)



from .stateful_chained_command import StatefulChainedCommand
from ...classes.processor import Processor
from ...partials.with_base.types import T_in, T_out


class StatefulChainedProcessor(Processor[T_in, T_out]):
    command_class = StatefulChainedCommand



from abc import abstractmethod
from typing import Any, Protocol

from .processor_facade import ProcessorFacade
from ..with_base.types import T_in, T_out
from ..with_base.stage_callable import StageCallableType


class CommandInterface(
    Protocol[T_in, T_out],
):
    def __init__(
        self,
        processor: ProcessorFacade[T_in, T_out],
        payload: T_in,
        stages: list[StageCallableType],
        *args: Any,
        **kwds: Any,
    ) -> None: ...

    @abstractmethod
    async def execute(self) -> T_out: ...

    async def _call_stage(
        self, payload: T_in, stage: StageCallableType, *args: Any, **kwds: Any
    ) -> T_in: ...



from abc import ABC
from typing import Self

from ..with_base.types import T_in, T_out
from ..with_base.stage_callable import StageCallableType
from .pipeline_interface_mixin import (
    PipelineInterfaceMixin as ImplementsPipelineInterface,
)


class PipelineMixin(
    ImplementsPipelineInterface[T_in, T_out],
    ABC,
):
    def pipe(self, stage: StageCallableType) -> Self:
        return self.add_item(stage)



from typing import Any, Optional, Protocol

from .processor_interface import ProcessorInterface
from ..with_base.types import T_in, T_out
from ..with_base.pipeline_interface import (
    PipelineInterface as WithPipelineBaseInterface,
)
from ..with_pipeline_as_list.pipeline_interface_mixin import (
    PipelineInterfaceMixin as WithPipelineAsListInterface,
)


class PipelineInterfaceMixin(
    WithPipelineAsListInterface,
    WithPipelineBaseInterface,
    Protocol[T_in, T_out],
):
    def __init__(
        self,
        processor: Optional[ProcessorInterface[T_in, T_out]] = None,
        *args: Any,
        **kwds: Any,
    ) -> None: ...

    def _make_processor(self) -> ProcessorInterface[T_in, T_out]: ...

    async def process(
        self, payload: T_in, *args: Any, **kwds: Any
    ) -> T_out: ...
